let the_type_constants = ref ["bool",bool_tyop; "fun",fun_tyop]

let types() = !the_type_constants

let get_type_arity s =
  match assoc s (!the_type_constants) with
  | Typrim(_,n) -> n
  | Tydefined(_,n,_) -> n

let new_type'(name,tyop) =
  if can get_type_arity name then
    failwith ("new_type: type "^name^" has already been declared")
  else the_type_constants := (name,tyop)::(!the_type_constants)

let new_type(name,arity) =
  new_type'(name,new_prim_type_op(name,arity))

let mk_type(name,args) =
  let tyop = try assoc name (!the_type_constants) with Failure _ ->
    failwith ("mk_type: type "^name^" has not been defined") in
    tyapp(tyop,args)

let the_term_constants =
  ref ["=",eq_term aty]

let constants() = !the_term_constants

let get_const_type s = type_of (assoc s (!the_term_constants))

let new_constant'(name,c) =
  if can get_const_type name then
    failwith ("new_constant: constant "^name^" has already been declared")
  else the_term_constants := (name,c)::(!the_term_constants) 

let new_constant(name,ty) =
  new_constant'(name,new_prim_const(name,ty))

let mk_const(name,theta) =
  let tm = try assoc name (!the_term_constants) with Failure _ ->
    failwith "mk_const: not a constant name" in
  inst_const(tm,theta)

  let the_context = ref empty_context;;

  let the_axioms = ref ([]:thm list)

  let axioms() = !the_axioms

  let new_axiom tm =
    if Pervasives.compare (type_of tm) bool_ty = 0 then
      let th,ctx = axiom_sequent (!the_context) tm in
       (the_axioms := th::(!the_axioms); the_context := ctx;
        th)
    else failwith "new_axiom: Not a proposition"

  let the_definitions = ref ([]:thm list)

  let definitions() = !the_definitions

  let new_basic_definition tm =
    let c,dth = new_defined_const tm in
    match c with
    | Const(name,_,_) ->
        new_constant'(name,c); the_definitions := dth::(!the_definitions); dth

  let new_basic_type_definition tyname (absname,repname) dth =
    if exists (can get_const_type) [absname; repname] then
      failwith "new_basic_type_definition: Constant(s) already in use" else
    let atyop,abs,rep,dth1,dth2 = new_defined_type_op tyname (absname,repname) dth in
    let _ = try new_type'(tyname,atyop)
            with Failure _ ->
                failwith "new_basic_type_definition: Type already defined" in
    let _ = new_constant'(absname,abs)
    and _ = new_constant'(repname,rep) in
    dth1,dth2

let mk_fun_ty ty1 ty2 = mk_type("fun",[ty1; ty2]);;
let bty = mk_vartype "B";;

let is_eq tm =
  match tm with
    Comb(Comb(Const("=",_,Prim),_),_) -> true
  | _ -> false;;

let mk_eq =
  let eq = mk_const("=",[]) in
  fun (l,r) ->
    try let ty = type_of l in
        let eq_tm = inst [ty,aty] eq in
        mk_comb(mk_comb(eq_tm,l),r)
    with Failure _ -> failwith "mk_eq";;

let aconv s t = alphaorder s t = 0;;

let equals_thm th th' = dest_thm th = dest_thm th';;

let the_core_definitions = the_definitions;;
